using System;
using System.Net;
using NUnit.Framework;

using Smdn.Collections;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientTests {
    [Test]
    public void TestConstruct()
    {
      Assert.AreEqual(new Uri("pop://localhost/"),
                      DefaultPropertyAssertion(new PopClient()).Profile.Authority,
                      "#1 authority");

      Assert.AreEqual(new Uri("pop://user;AUTH=+APOP@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient(new Uri("pop://user;AUTH=+APOP@pop.example.net:10110/"))).Profile.Authority,
                      "#2 authority");

      Assert.AreEqual(new Uri("pop://pop.example.net/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net")).Profile.Authority,
                      "#3 authority");

      Assert.AreEqual(new Uri("pop://user@pop.example.net/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", "user")).Profile.Authority,
                      "#4 authority");

      Assert.AreEqual(new Uri("pops://user@pop.example.net/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", true, "user")).Profile.Authority,
                      "#5 authority");

      Assert.AreEqual(new Uri("pop://pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110)).Profile.Authority,
                      "#6 authority");

      Assert.AreEqual(new Uri("pop://user@pop.example.net:10110/"), 
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user")).Profile.Authority,
                      "#7 authority");

      Assert.AreEqual(new Uri("pop://user@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user")).Profile.Authority,
                      "#8 authority");

      Assert.AreEqual(new Uri("pops://user@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user")).Profile.Authority,
                      "#9 authority");

      Assert.AreEqual(new Uri("pop://user;AUTH=+APOP@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user", "+APOP")).Profile.Authority,
                      "#10 authority");

      Assert.AreEqual(new Uri("pops://user;AUTH=DIGEST-MD5@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user", "DIGEST-MD5")).Profile.Authority,
                      "#11 authority");

      Assert.AreEqual(new Uri("pops://user@pop.example.net:10110/"),
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user", null, 1000), 1000).Profile.Authority,
                      "#12 authority");

      var profile = new PopClientProfile(new Uri("pops://user@pop.example.net/"));

      Assert.AreEqual(new Uri("pops://user@pop.example.net/"),
                      DefaultPropertyAssertion(new PopClient(profile)).Profile.Authority,
                      "#13 authority");
    }

    private PopClient DefaultPropertyAssertion(PopClient client)
    {
      return DefaultPropertyAssertion(client, System.Threading.Timeout.Infinite);
    }

    private PopClient DefaultPropertyAssertion(PopClient client, int timeoutMilliseconds)
    {
      Assert.IsNotNull(client.Profile);
      Assert.IsTrue(client.Profile.UseTlsIfAvailable);
      Assert.IsFalse(client.DeleteAfterRetrieve);
      Assert.AreEqual(timeoutMilliseconds, client.Timeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.ReceiveTimeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.SendTimeout);

      TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(delegate{ Assert.AreEqual(0L, client.TotalSize); });
      TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(delegate{ Assert.AreEqual(0L, client.MessageCount); });

      return client;
    }

    [Test]
    public void TestConnectWithPassword()
    {
      Connect(delegate(PopPseudoServer server, PopClient client) {
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    [Test]
    public void TestConnectWithCredentials()
    {
      Connect(delegate(PopPseudoServer server, PopClient client) {
        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    private void Connect(Action<PopPseudoServer, PopClient> action)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        using (var client = new PopClient(server.Host, server.Port, "user")) {
          Assert.IsFalse(client.IsConnected);
          TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(() => Assert.IsNotNull(client.ServerCapabilities));

          // greeting
          server.EnqueueResponse("+OK\r\n");
          // CAPA
          server.EnqueueResponse("+OK\r\n" +
                                 ".\r\n");
          // USER/PASS
          server.EnqueueResponse("+OK\r\n");
          server.EnqueueResponse("+OK\r\n");
          // STAT
          server.EnqueueResponse("+OK 3 360\r\n");

          action(server, client);

          Assert.AreEqual(3L, client.MessageCount);
          Assert.AreEqual(360L, client.TotalSize);
        }
      }
    }

    [Test]
    public void TestDisconnect()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.Disconnect();

        Assert.IsFalse(client.IsConnected);
        TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      });
    }

    [Test]
    public void TestCheckConnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.Disconnect();

        Assert.IsFalse(client.IsConnected);

        TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(delegate { client.KeepAlive(); });
      });
    }

    [Test]
    public void TestLogout()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // QUIT
        server.EnqueueResponse("+OK\r\n");

        client.Logout();

        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");

        Assert.IsFalse(client.IsConnected);
        TestUtils.ExpectExceptionThrown<PopProtocolViolationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      });
    }

    [Test]
    public void TestReconnect()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // QUIT
        server.EnqueueResponse("+OK\r\n");

        client.Logout();

        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");

        Assert.IsFalse(client.IsConnected);

        // reconnect
        server.Stop();
        server.Start();

        client.Profile.Host = server.Host;
        client.Profile.Port = server.Port;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER/PASS
        server.EnqueueResponse("+OK\r\n");
        server.EnqueueResponse("+OK\r\n");
        // STAT
        server.EnqueueResponse("+OK 0 0\r\n");

        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    [Test]
    public void TestGetMessage()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        // get again
        Assert.AreSame(message, client.GetMessage(1L));
      });
    }

    [Test]
    public void TestGetMessageGetUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");
        // UIDL
        server.EnqueueResponse("+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = client.GetMessage(1L, true);

        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "UIDL 1\r\n");

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test, ExpectedException(typeof(PopMessageDeletedException))]
    public void TestGetMessageDeleted()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        // DELE
        server.EnqueueResponse("+OK\r\n");

        message.MarkAsDeleted();

        // throw exception
        client.GetMessage(1L);
      });
    }

    [Test, ExpectedException(typeof(PopMessageNotFoundException))]
    public void TestGetMessageNotFound()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("-ERR no such message\r\n");

        // throws exception
        client.GetMessage(1L);
      });
    }


    [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
    public void TestGetMessageZero()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.GetMessage(0L);
      });
    }

    [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
    public void TestGetMessageGreaterThanMessageCount()
    {
      TestUtils.TestAuthenticated(3, delegate(PopPseudoServer server, PopClient client) {
        client.GetMessage(4L);
      });
    }

    [Test]
    public void TestGetMessageByUid()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // UIDL
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        // LIST
        server.EnqueueResponse("+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.AreEqual(server.DequeueRequest(), "UIDL\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 2\r\n");

        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreEqual(240L, message.Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", message.UniqueId);
      });
    }

    [Test, ExpectedException(typeof(PopMessageDeletedException))]
    public void TestGetMessageByUidDeleted()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // UIDL
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        // LIST
        server.EnqueueResponse("+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        // DELE
        server.EnqueueResponse("+OK\r\n");

        message.MarkAsDeleted();

        // throw exception
        client.GetMessage(message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageByUidAlreadyListed()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        // UIDL
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");
        Assert.AreEqual(server.DequeueRequest(), "UIDL\r\n");

        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.IsNotNull(message);
        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreSame(messages[1], message);
      });
    }

    [Test, ExpectedException(typeof(ArgumentNullException))]
    public void TestGetMessageByUidNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.GetMessage((string)null);
      });
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestGetMessageByUidEmpty()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.GetMessage(string.Empty);
      });
    }

    [Test, ExpectedException(typeof(PopMessageNotFoundException))]
    public void TestGetMessageByUidNotFound()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // UIDL
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        client.GetMessage("non-existent");
      });
    }

    [Test]
    public void TestGetMessages()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messages = client.GetMessages().ToArray();

        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);

        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messagesSecond = client.GetMessages().ToArray();

        Assert.AreSame(messages[0], messagesSecond[0]);
        Assert.AreSame(messages[1], messagesSecond[1]);
        Assert.AreSame(messages[2], messagesSecond[2]);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        // UIDL
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestGetMessagesEmpty()
    {
      GetMessagesEmpty(false);
    }

    [Test]
    public void TestGetMessagesEmptyGetUniqueId()
    {
      GetMessagesEmpty(true);
    }

    private void GetMessagesEmpty(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");

        var messages = (getUniqueId ? client.GetMessages(true) : client.GetMessages()).ToArray();

        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

        Assert.AreEqual(0, messages.Length);
      });
    }

    [Test]
    public void TestCancelDelete()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        server.DequeueRequest();

        // DELE
        server.EnqueueResponse("+OK\r\n");

        message.MarkAsDeleted();

        server.DequeueRequest();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        // RSET
        server.EnqueueResponse("+OK\r\n");

        client.CancelDelete();

        Assert.AreEqual(server.DequeueRequest(), "RSET\r\n");

        Assert.IsFalse(message.IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestCancelDeleteMessageNotListed()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // RSET
        server.EnqueueResponse("+OK\r\n");

        client.CancelDelete();

        Assert.AreEqual(server.DequeueRequest(), "RSET\r\n");
      });
    }

    [Test]
    public void TestKeepAlive()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // NOOP
        server.EnqueueResponse("+OK\r\n");

        client.KeepAlive();

        Assert.AreEqual(server.DequeueRequest(), "NOOP\r\n");
      });
    }

    [Test, ExpectedException(typeof(Smdn.Net.Pop3.Protocol.PopConnectionException))]
    public void TestKeepAliveDisconnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.Stop();

        client.KeepAlive();
      });
    }
  }
}
